﻿using HeuristicLab.Algorithms.PushGP;
using HeuristicLab.Algorithms.PushGP.Expressions;
using HeuristicLab.Algorithms.PushGP.Generators;
using HeuristicLab.Algorithms.PushGP.Stack;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace HeuristicLab.Tests.Interpreter.Expressions
{
    [TestClass]
    public class CodeExpressionTests : CommonTests<Expression>
    {
        protected override string TypeName { get { return "CODE"; } }
        protected override IStack<Expression> Stack { get { return interpreter.CodeStack; } }

        private CodeGenerator cg;
        public CodeExpressionTests()
        {
            this.cg = new CodeGenerator(interpreter);
        }

        [TestMethod]
        public void TestAppendLists()
        {
            var first = Parser.Parse("( A )");
            var second = Parser.Parse("( B )");
            var result = Parser.Parse("( A B )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeAppendExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestAppendListAndLiteral()
        {
            var first = Parser.Parse("( A )");
            var second = Parser.Parse("5");
            var result = Parser.Parse("( A 5 )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeAppendExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestAppendLiterals()
        {
            var first = Parser.Parse("4");
            var second = Parser.Parse("5");
            var result = Parser.Parse("( 4 5 )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeAppendExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestAtomList()
        {
            var first = Parser.Parse("( A )");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeAtomExpression());

            Assert.AreEqual(false, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestAtomLiteral()
        {
            var first = Parser.Parse("5");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeAtomExpression());

            Assert.AreEqual(true, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestAtomSingleInstruction()
        {
            var first = Parser.Parse("INTEGER.+");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeAtomExpression());

            Assert.AreEqual(true, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestCarList()
        {
            var first = Parser.Parse("( A B C )");
            var result = Parser.Parse("A");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeCarExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestCarNoList()
        {
            var first = Parser.Parse("5");
            var result = Parser.Parse("5");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeCarExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestCdrList()
        {
            var first = Parser.Parse("( A B C )");
            var result = Parser.Parse("( B C )");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestCdrNoList()
        {
            var first = Parser.Parse("5");
            var result = Parser.Parse("( )");

            interpreter.CodeStack.Push(first);
            interpreter.Interpret(new CodeCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestCons()
        {
            var first = Parser.Parse("( B C )");
            var second = Parser.Parse("A");
            var result = Parser.Parse("( A B C )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeConsExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestContainsTrue()
        {
            var first = Parser.Parse("A");
            var second = Parser.Parse("( A B C )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeContainsExpression());

            Assert.AreEqual(true, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestContainsFalse()
        {
            var first = Parser.Parse("D");
            var second = Parser.Parse("( A B C )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeContainsExpression());

            Assert.AreEqual(false, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestContainer()
        {
            var first = Parser.Parse("( A )");
            var second = Parser.Parse("( B ( C ( A ) ) ( D ( A ) ) )");
            var result = Parser.Parse("( C ( A ) )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeContainerExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestDefinition()
        {
            var code = Parser.Parse("( A )");

            interpreter.CustomExpressions.Add("Test", code);
            interpreter.NameStack.Push("Test");

            interpreter.Interpret(new CodeDefinitionExpression());

            Assert.AreEqual(code, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestDefinitionWithoutMatchingName()
        {
            var code = Parser.Parse("( A )");
            var program = new CodeDefinitionExpression();

            interpreter.CustomExpressions.Add("Test1", code);
            interpreter.NameStack.Push("Test2");
            interpreter.Interpret(program);

            this.TestStackCounts(nameStack: 1);
        }

        [TestMethod]
        public void TestDiscrepancy()
        {
            var first = Parser.Parse("( A B )");
            var second = Parser.Parse("( B C )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeDiscrepancyExpression());

            Assert.AreEqual(2, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDiscrepancyAllEqual()
        {
            var first = Parser.Parse("( A B )");
            var second = Parser.Parse("( B A )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeDiscrepancyExpression());

            Assert.AreEqual(0, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDiscrepancyComplex()
        {
            var first = Parser.Parse("( A A B C D A )");
            var second = Parser.Parse("( A A B D E )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeDiscrepancyExpression());

            Assert.AreEqual(3, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDiscrepancWithLists()
        {
            var first = Parser.Parse("( ( A ) A ( B ) C D A )");
            var second = Parser.Parse("( A ( A ) B D E )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeDiscrepancyExpression());

            Assert.AreEqual(5, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDo()
        {
            var prog = Parser.Parse("( 10 )");

            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoExpression());

            Assert.AreEqual(10, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDoX()
        {
            var prog = Parser.Parse("( 10 )");

            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoXExpression());

            Assert.AreEqual(10, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDoCount()
        {
            var prog = Parser.Parse("INTEGER.+");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoCountExpression());

            Assert.AreEqual(6, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDoCountWithNegativeCounter()
        {
            var prog = Parser.Parse("INTEGER.+");

            interpreter.IntegerStack.Push(-1);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoCountExpression());

            Assert.AreEqual(-1, interpreter.IntegerStack.Top);
            this.TestStackCounts(codeStack: 1, integerStack: 1);
        }

        [TestMethod]
        public void TestDoRange()
        {
            var prog = Parser.Parse("INTEGER.+");

            interpreter.IntegerStack.Push(3, 5);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoRangeExpression());

            Assert.AreEqual(12, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDoRangeWithEqualIndecators()
        {
            var prog = Parser.Parse("INTEGER.+");

            interpreter.IntegerStack.Push(3, 3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoRangeExpression());

            Assert.AreEqual(3, interpreter.IntegerStack.Top);
            this.TestStackCounts(codeStack: 1, integerStack: 2);
        }

        [TestMethod]
        public void TestDoRangeWithNegativeIndecators()
        {
            var prog = Parser.Parse("INTEGER.+");

            interpreter.IntegerStack.Push(-3, -5);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoRangeExpression());

            Assert.AreEqual(-12, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDoTimes()
        {
            var prog = Parser.Parse("( INTEGER.DUP INTEGER.+ )");

            interpreter.IntegerStack.Push(2, 3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoTimesExpression());

            Assert.AreEqual(32, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestDoTimesWithNegativeCounter()
        {
            var prog = Parser.Parse("( INTEGER.DUP INTEGER.+ )");

            interpreter.IntegerStack.Push(2, -3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeDoTimesExpression());

            Assert.AreEqual(-3, interpreter.IntegerStack.Top);
            this.TestStackCounts(codeStack: 1, integerStack: 2);
        }

        [TestMethod]
        public void TestNestedDoRange()
        {
            interpreter.Interpret("( 0 2 CODE.QUOTE ( 1 INTEGER.+ 0 3 CODE.QUOTE ( 1 INTEGER.+ INTEGER.* ) CODE.DO*RANGE INTEGER.+ ) CODE.DO*RANGE )");

            Assert.AreEqual(144, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestNestedDoCount()
        {
            interpreter.Interpret("( 2 CODE.QUOTE ( 1 INTEGER.+ 3 CODE.QUOTE ( 1 INTEGER.+ INTEGER.* ) CODE.DO*COUNT INTEGER.+ ) CODE.DO*COUNT )");

            Assert.AreEqual(144, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestNestedDoTimes()
        {
            interpreter.Interpret("( 3 CODE.QUOTE ( 2 3 CODE.QUOTE ( 2 INTEGER.* ) CODE.DO*TIMES INTEGER.+ ) CODE.DO*TIMES )");

            Assert.AreEqual(128, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestExtract()
        {
            var prog = Parser.Parse("( A ( B C ( D ) E ) F )");
            var result = Parser.Parse("( D )");

            interpreter.IntegerStack.Push(5);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeExtractExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestExtractInteger()
        {
            var prog = Parser.Parse("( A ( 5 C ( 4.2 ) INTEGER.+ ) FALSE )");
            var result = Parser.Parse("5");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeExtractExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestExtractFloat()
        {
            var prog = Parser.Parse("( A ( 5 C ( 4.2 ) INTEGER.+ ) FALSE )");
            var result = Parser.Parse("4.2");

            interpreter.IntegerStack.Push(6);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeExtractExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestExtractBoolean()
        {
            var prog = Parser.Parse("( A ( 5 C ( 4.2 ) INTEGER.+ ) FALSE )");
            var result = Parser.Parse("FALSE");

            interpreter.IntegerStack.Push(8);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeExtractExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestExtractInstruction()
        {
            var prog = Parser.Parse("( A ( 5 C ( 4.2 ) INTEGER.+ ) FALSE )");
            var result = Parser.Parse("INTEGER.+");

            interpreter.IntegerStack.Push(7);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeExtractExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestFromBoolean()
        {
            var result = Parser.Parse("FALSE");
            interpreter.BooleanStack.Push(false);
            interpreter.Interpret(new CodeFromBooleanExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestFromFloat()
        {
            var result = Parser.Parse("4.1");
            interpreter.FloatStack.Push(4.1);
            interpreter.Interpret(new CodeFromFloatExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestFromInteger()
        {
            var result = Parser.Parse("4");
            interpreter.IntegerStack.Push(4);
            interpreter.Interpret(new CodeFromIntegerExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestFromName()
        {
            var result = Parser.Parse("A");
            interpreter.NameStack.Push("A");
            interpreter.Interpret(new CodeFromNameExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestIfTrue()
        {
            var first = Parser.Parse("FALSCH");
            var second = Parser.Parse("WAHR");

            interpreter.BooleanStack.Push(true);
            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeIfExpression());

            Assert.AreEqual("WAHR", interpreter.NameStack.Top);
            this.TestStackCounts(nameStack: 1);
        }

        [TestMethod]
        public void TestIfFalse()
        {
            var first = Parser.Parse("FALSCH");
            var second = Parser.Parse("WAHR");

            interpreter.BooleanStack.Push(false);
            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeIfExpression());

            Assert.AreEqual("FALSCH", interpreter.NameStack.Top);
            this.TestStackCounts(nameStack: 1);
        }

        [TestMethod]
        public void TestInsert()
        {
            var first = Parser.Parse("( A ( B ) C D )");
            var second = Parser.Parse("( E )");
            var result = Parser.Parse("( A ( B ) ( E ) D )");

            interpreter.IntegerStack.Push(2);
            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeInsertExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestInsertWithNegativeN()
        {
            var first = Parser.Parse("( A ( B ) C D )");
            var second = Parser.Parse("( E )");
            var result = Parser.Parse("( A ( B ) ( E ) D )");

            interpreter.IntegerStack.Push(-2);
            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeInsertExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestInsertWithTooBigN()
        {
            var first = Parser.Parse("( A ( B ) C D )");
            var second = Parser.Parse("( E )");
            var result = Parser.Parse("( A ( B ) ( E ) D )");

            interpreter.IntegerStack.Push(10);
            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeInsertExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestList()
        {
            var first = Parser.Parse("A");
            var second = Parser.Parse("( B )");
            var result = Parser.Parse("( A ( B ) )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeListExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestMemberTrue()
        {
            var first = Parser.Parse("( A B ( B ) C ");
            var second = Parser.Parse("( B )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeMemberExpression());

            Assert.AreEqual(true, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestMemberFalse()
        {
            var first = Parser.Parse("( A B C ");
            var second = Parser.Parse("( B )");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeMemberExpression());

            Assert.AreEqual(false, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestMemberIfNoList()
        {
            var first = Parser.Parse("B");
            var second = Parser.Parse("B");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodeMemberExpression());

            Assert.AreEqual(true, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestNoop()
        {
            interpreter.Interpret(new CodeNoopExpression());

            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthIfNoList()
        {
            var prog = Parser.Parse("A");
            var result = Parser.Parse("A");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthIfEmpty()
        {
            var prog = Parser.Parse("( )");
            var result = Parser.Parse("( )");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNth()
        {
            var prog = Parser.Parse("( A B C D E )");
            var result = Parser.Parse("D");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthWithTooBigN()
        {
            var prog = Parser.Parse("( A B C D E )");
            var result = Parser.Parse("D");

            interpreter.IntegerStack.Push(13);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthWithNegativeN()
        {
            var prog = Parser.Parse("( A B C D E )");
            var result = Parser.Parse("D");

            interpreter.IntegerStack.Push(-3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthCdr()
        {
            var prog = Parser.Parse("( A B C D E )");
            var result = Parser.Parse("D");

            interpreter.IntegerStack.Push(2);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthCdrWithTooBigN()
        {
            var prog = Parser.Parse("( A B C D E )");
            var result = Parser.Parse("D");

            interpreter.IntegerStack.Push(10);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthCdrWithNegativeN()
        {
            var prog = Parser.Parse("( A B C D E )");
            var result = Parser.Parse("D");

            interpreter.IntegerStack.Push(-2);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthCdrIfEmpty()
        {
            var prog = Parser.Parse("( )");
            var result = Parser.Parse("( )");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNthCdrIfNoList()
        {
            var prog = Parser.Parse("A");
            var result = Parser.Parse("A");

            interpreter.IntegerStack.Push(3);
            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNthCdrExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestNullTrue()
        {
            var prog = Parser.Parse("( )");

            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNullExpression());

            Assert.AreEqual(true, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestNullFalse()
        {
            var prog = Parser.Parse("A");

            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeNullExpression());

            Assert.AreEqual(false, interpreter.BooleanStack.Top);
            this.TestStackCounts(booleanStack: 1);
        }

        [TestMethod]
        public void TestPosition()
        {
            var first = Parser.Parse("( A B C D )");
            var second = Parser.Parse("C");

            interpreter.CodeStack.Push(second, first);
            interpreter.Interpret(new CodePositionExpression());

            Assert.AreEqual(2, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestQuote()
        {
            var result = Parser.Parse("A");

            interpreter.Interpret("( CODE.QUOTE A )");

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        [TestMethod]
        public void TestSize()
        {
            var prog = Parser.Parse("( A ( B ( C ) ) ( D ) INTEGER.+ )");

            interpreter.CodeStack.Push(prog);
            interpreter.Interpret(new CodeSizeExpression());

            Assert.AreEqual(4, interpreter.IntegerStack.Top);
            this.TestStackCounts(integerStack: 1);
        }

        [TestMethod]
        public void TestSubstitude()
        {
            var first = Parser.Parse("( A B C D A B C D )");
            var second = Parser.Parse("E");
            var third = Parser.Parse("A");
            var result = Parser.Parse("( E B C D E B C D )");

            interpreter.CodeStack.Push(third, second, first);
            interpreter.Interpret(new CodeSubstitutionExpression());

            Assert.AreEqual(result, interpreter.CodeStack.Top);
            this.CheckOtherStacksAreEmpty();
        }

        protected override Expression[] GetValues(int count)
        {
            var values = new Expression[count];

            for (var i = 0; i < count; i++)
            {
                values[i] = new CodeNoopExpression();
            }

            return values;
        }

        protected override Expression[] Get2Different()
        {
            return new Expression[]
            {
                new CodeNoopExpression(),
                new CodeNullExpression()
            };
        }

        protected override void CheckOtherStacksAreEmpty()
        {
            this.TestStackCounts(codeStack: null);
        }
    }
}
